'''
Graphics for the world where the things interact, live, and die.
But since some sensory input is implemented using Tkinter methods,
the Canvas objects are in a sense part of the things themselves,
and graphics cannot be completely turned off without interfering with
senses.
'''

from Tkinter import *
from thing import *

import time

VERSION = 1.0

# Dimensions of the world window
WIDTH = 450
HEIGHT = 450

# Delay in microseconds between steps during Run when it's being displayed
# (usually much more than this)
STEP_DELAY = 50

# Wall color
BRICK = "#bb4422"
# Thickness of all walls
WALL_THICKNESS = 8

WORLD_FILE = 'world'

THING_SIDE = 12

def read_data(data_file):
    """Read in various data from data_file."""
    dct = {}
    if not data_file:
        data_file = WORLD_FILE
    print 'Loading world data from', data_file
    fl = open(data_file)
    raw = fl.read().split('\n\n')
    for line in raw:
        split_eq = line.split('=')
        if len(split_eq) > 1:
            key, value = split_eq
            # key could includes lines of comments
            # get rid of everything after the last \n
            key = key.rpartition('\n')[2]
            dct[key.strip()] = eval(value, globals(), dct)
    return dct

class WorldTop(Toplevel):
    '''Toplevel window to hold the WorldFrame when it's not under root.'''

    def __init__(self, width=WIDTH, height=HEIGHT):
        Toplevel.__init__(self)
        self.frame = WorldFrame(self, width, height)

class WorldFrame(Frame):
    '''Frame containing the World canvas where the world is drawn, and some buttons and menus.'''

    # Starting value for whether all steps are displayed during Run
    showing_steps = False
    showing_genomes = False
    
    def __init__(self, parent, width=WIDTH, height=HEIGHT, worldfile=None):
        '''Creates the Canvas where everything happens and the menus and buttons.'''
        Frame.__init__(self, parent)
        self.grid(row = 0)
        parent.title("The World")

        self.canvas = World(self, width=width, height=height, worldfile=worldfile)

        ## Menus, at the top
        # Menu for controlling type of evolution
        adapt_var = IntVar()
        adapt_var.set(DONT_EVOLVE)
        self.evolmenu = Menubutton(self, text = "Learn/Evolve")
        self.evolmenu.grid(row = 0, column = 0)
        self.evolmenu.menu = Menu(self.evolmenu)
        self.evolmenu["menu"] = self.evolmenu.menu
        self.evolmenu.menu.add_radiobutton(label = "Q-learn",
                                           command = self.learn, variable = adapt_var, value = LEARN)
        self.evolmenu.menu.add_radiobutton(label = "No learning or reproduction",
                                           command = self.unrepro, variable = adapt_var, value = DONT_EVOLVE)
        self.evolmenu.menu.add_radiobutton(label = "Copy without evolving",
                                           command = self.unevolve, variable = adapt_var, value = DUMB_COPYING)
        self.evolmenu.menu.add_radiobutton(label = "Evolve by copying",
                                           command = self.copy, variable = adapt_var, value = COPYING)
        self.evolmenu.menu.add_radiobutton(label = "Evolve by mating",
                                           command = self.mate, variable = adapt_var, value = MATING)

        # Menu for controlling how many steps happen during Run
        self.run_length = IntVar()
        self.run_length.set(100)
        self.runmenu = Menubutton(self, text = "Run length")
        self.runmenu.grid(row = 0, column = 1)
        self.runmenu.menu = Menu(self.runmenu)
        self.runmenu["menu"] = self.runmenu.menu
        self.runmenu.menu.add_radiobutton(label = "100 steps", variable = self.run_length, value = 100)
        self.runmenu.menu.add_radiobutton(label = "200 steps", variable = self.run_length, value = 200)
        self.runmenu.menu.add_radiobutton(label = "500 steps", variable = self.run_length, value = 500)
        self.runmenu.menu.add_radiobutton(label = "1000 steps", variable = self.run_length, value = 1000)
        self.runmenu.menu.add_radiobutton(label = "1500 steps", variable = self.run_length, value = 1500)

        # Menu for controlling whether senses are displayed and all steps are displayed during Run
        self.viewmenu = Menubutton(self, text = "View")
        self.viewmenu.grid(row = 0, column = 2)
        self.viewmenu.menu = Menu(self.viewmenu)
        self.viewmenu["menu"] = self.viewmenu.menu
        self.viewmenu.menu.add_command(label = "Show senses", command = self.toggle_show_senses)
        self.viewmenu.menu.add_command(label = "Always update graphics",
                                       command = self.toggle_show_steps)
        self.viewmenu.menu.add_command(label = "Show genomes after Run",
                                       command = self.toggle_show_genomes)

        ## Buttons, at the bottom
        # Reinitialize (doesn't really work now)
        self.button0 = Button(self, text = "Reinit")
        self.button0.grid(row = 2, column = 0)
        self.button0.bind("<Button-1>", self.canvas.reinit)
        # One step of the world
        self.button1 = Button(self, text = "Step")
        self.button1.grid(row = 2, column = 1)
        self.button1.bind("<Button-1>", self.canvas.step)
        # Multiple steps of the world (number controlled by Run length menu)
        self.button2 = Button(self, text = "Run")
        self.button2.grid(row = 2, column = 2)
        self.button2.bind("<Button-1>", self.canvas.start_run)

    def learn(self):
        '''Called when "Q-learn" is selected.'''
        self.canvas.evolve = False
        self.canvas.learn = True
        self.canvas.reproduction = LEARN

    def copy(self):
        '''Called when "Reproduce by copying" is selected.'''
        self.canvas.evolve = True
        self.canvas.learn = False
        self.canvas.reproduction = COPYING

    def mate(self):
        '''Called when "Reproduce by mating" is selected.'''
        self.canvas.evolve = True
        self.canvas.learn = False
        self.canvas.reproduction = MATING

    def unevolve(self):
        """Called when "Copy without evolving" is selected."""
        self.canvas.evolve = False
        self.canvas.reproduction = DUMB_COPYING

    def unrepro(self):
        """Called when "No reproduction" is selected."""
        self.canvas.evolve = False
        self.canvas.learn = False
        self.canvas.reproduction = NO_REPRO
    
    def toggle_show_senses(self):
        '''Toggle whether senses are displayed.'''
        if self.canvas.show_senses:
            self.viewmenu.menu.entryconfigure(0, label = "Show senses")
            self.canvas.toggle_show_senses(False)
        else:
            self.viewmenu.menu.entryconfigure(0, label = "Hide senses")
            self.canvas.toggle_show_senses(True)

    def toggle_show_steps(self):
        '''Toggle whether all steps are displayed during Run.'''
        if self.showing_steps:
            self.showing_steps = False
            self.viewmenu.menu.entryconfigure(1, label = "Always update graphics")
        else:
            self.showing_steps = True
            self.viewmenu.menu.entryconfigure(1, label = "Update graphics after Run")

    def toggle_show_genomes(self):
        '''Toggle whether genomes are showed after Run.'''
        if self.showing_genomes:
            self.showing_genomes = False
            self.viewmenu.menu.entryconfigure(2, label = "Show genomes after Run")
        else:
            self.showing_genomes = True
            self.viewmenu.menu.entryconfigure(2, label = "Don't show genomes after Run")

class World(Canvas):
    '''Canvas to display everything in the world.'''

    # Min and max coordinates for things
    min_x = 5 + THING_SIDE + WALL_THICKNESS
    max_x = WIDTH - THING_SIDE - WALL_THICKNESS
    min_y = 5 + THING_SIDE + WALL_THICKNESS
    max_y = HEIGHT - THING_SIDE - WALL_THICKNESS
    
    # Default for whether senses are displayed
    show_senses = False

    def __init__(self, parent, width=WIDTH, height=HEIGHT, worldfile=None):
        '''Initialize a bunch of attributes, create the things and walls.'''
        Canvas.__init__(self, parent, width=width, height=height, bg="black")
        self.parent = parent
        self.grid(row = 1, columnspan = 3)

        # Read in features from file
        self.features = read_data(worldfile)
        
        # Thing and wall attributes
        self.things = {}
        self.clusters = {}
        self.sounds = []
        self.lights = []
        self.walls = []
        self.vert_wall_bounds = []
        self.vert_real_wall_bounds = []
        self.horiz_wall_bounds = []
        self.horiz_real_wall_bounds = []

        # Control how and whether evolution or learning happens
        self.learn = False
        self.evolve = False
        self.reproduction = NO_REPRO

        # For statistics only
        self.starved = 0
        self.died_age = 0
        self.copied = 0
        self.mated = 0

        # Run control
        self.running = False
        self.abort_run = False
        self.run_time = 0.0

        # Reproduction related parameter
        self.repro_aging = self.features.get('REPRO_AGING', 100)

        # Create walls
        self.add_wall(Wall(self, True, exact = (7, 0, height),
                           world_width = width, world_height = height), True)
        self.add_wall(Wall(self, True, exact = (width, 0, height),
                           world_width = width, world_height = height), True)
        self.add_wall(Wall(self, False, exact = (7, 0, width),
                           world_width = width, world_height = height), False)
        self.add_wall(Wall(self, False, exact = (height, 0, width),
                           world_width = width, world_height = height), False)
        v_walls = self.features.get('V_WALLS', [])
        h_walls = self.features.get('H_WALLS', [])
        for w in range(len(v_walls)):
            self.add_wall(Wall(self, True, v_walls[w][0], v_walls[w][1], v_walls[w][2],
                               world_width = width, world_height = height), True)
        for w in range(len(h_walls)):
            self.add_wall(Wall(self, False, h_walls[w][0], h_walls[w][1], h_walls[w][2],
                               world_width = width, world_height = height), False)
        # Create things
        for tp, dct in self.features['THINGS'].iteritems():
            if "cluster" in dct:
                self.clusters[tp] = self.get_cluster_centers(dct["cluster"][0])
            for t in range(dct["init"]):
                self.add_thing(tp, dct)
        # Create lights
        for light_dict in self.features.get('LIGHTS', []):
            self.add_light(light_dict['coords'], light_dict['color'])

        # Number of time steps run
        self.n_steps = 0
        # Things to be copied at the end of this time step
        self.copy = []
        # Pairs of critters to be mated at the end of this time step
        self.mates = []
        # Number of critters to be born on the next time step
#        self.born = []

    def run_length(self):
        '''Number of steps on a given Run.'''
        return self.parent.run_length.get()

    def add_wall(self, wall, vert):
        """Add a wall to the world"""
        self.walls.append(wall)
        if vert:
            self.vert_wall_bounds.append(wall.get_bounds())
            self.vert_real_wall_bounds.append(wall.get_bounds(False))
        else:
            self.horiz_wall_bounds.append(wall.get_bounds())
            self.horiz_real_wall_bounds.append(wall.get_bounds(False))

    def add_sound(self, coords, text):
        '''Add a new sound at the given position with the given sound.'''
        self.sounds.append(Sound(self, coords, text))

    def add_light(self, coords, color):
        self.lights.append(Light(self, coords, color))

    def cluster_center_obj(self, coords):
        '''A little circle showing where a cluster of things is.'''
        self.create_oval(coords[0] - 2, coords[1] - 1, coords[0] + 1, coords[1] + 2,
                         width = 1, outline = "green", fill = "green")

    def get_thing_prop(self, kind, prop):
        '''Value for a property of things of type kind.'''
        dct = self.features['THINGS'].get(kind)
        if dct:
            return dct.get(prop, None)

    def add_thing(self, kind, props):
        """Add a thing to the world."""
        maximum = (self.get_thing_prop(kind, 'max') or 1000)
        if self.n_things(kind) >= maximum:
            pass
        else:
            thing = kind(self, coords=self.get_thing_pos(props, kind), child=False)
            self.things[thing.id] = thing

    def n_things(self, kind):
        '''Number of things of type kind.'''
        return len([t for t in self.things.values() if isinstance(t, kind)])

    def reinit(self, event):
        """Reinitialize the world, creating minimum numbers of things"""
        for tp, dct in self.features['THINGS'].iteritems():
            for t in range(self.get_thing_n(tp.__name__), dct["init"]):
                self.add_thing(tp, dct)
        self.n_steps = 0

    def reached_max(self, kind):
        '''Maximum number of things of kind.'''
        maximum = self.get_thing_prop(kind, 'max')
        return not maximum or self.n_things(kind) >= maximum

    def regenerate(self):
        """Recreate the missing things in the world, changing min if specified for a type."""
#        for props in self.born:
#            self.add_thing(props)
#        self.born = []

        for tp, dct in self.features['THINGS'].iteritems():
            if 'change' in dct and self.n_steps != 0:
                change_props = dct['change']
                freq = change_props.get('freq', 10)
                if self.n_steps % freq == 0:
                    # Change 'min'
                    incr = change_props.get('incr', 1)
                    if incr > 0:
                        new_min = dct['min'] + incr
                        if dct.get('max', 1000) > new_min:
                            print 'Changing min for', tp.__name__, 'to', new_min
                            dct['min'] = new_min
                    else:
                        new_min = dct['min'] + incr
                        if new_min > dct.get('absmin', 0):
                            print 'Changing min for', tp.__name__, 'to', new_min
                            dct['min'] = new_min
            if "min" in dct:
                for t in range(self.get_thing_n(tp.__name__), dct["min"]):
                    self.add_thing(tp, dct)

            if 'growth' in dct and not self.reached_max(tp):
                for t in range(dct['growth']):
                    self.add_thing(tp, dct)

    def reproduce(self):
        '''Copy or mate things on the "copy" and "mate" lists.'''
        copy_string = ''
        for t in self.copy:
            if not self.reached_max(type(t)):
                thing = t.copy()
                self.things[thing.id] = thing
                self.copied += 1
        self.copy = []
        for pair in self.mates:
            if not self.reached_max(type(pair[0])):
                self.mate(pair[0], pair[1])
                self.mated += 1
        self.mates = []

    def mate(self, critter1, critter2):
        '''Create two new offspring using crossover.'''
        genome1 = critter1.genome
        genome2 = critter2.genome
        offspring1 = critter1.copy()
        offspring2 = critter2.copy()
        crossover_point = random.randint(1, genome1.length)
        if random.random() < Genome.CROSSOVER:
            offspring1.genome.crossover_right(genome2, crossover_point)
            offspring2.genome.crossover_right(genome1, crossover_point)
        self.things[offspring1.id] = offspring1
        self.things[offspring2.id] = offspring2
        critter1.get_older(self.repro_aging)
        critter2.get_older(self.repro_aging)

    def get_cluster_centers(self, n):
        """Find a new set of n cluster centers."""
        cs = []
        for i in range(n):
            c = self.get_vacant_pos()
            cs.append(c)
        return cs

    def get_vacant_pos(self):
        """Find a position that doesn't overlap with any walls."""
        coords = (random.randint(self.min_x, self.max_x),
                  random.randint(self.min_y, self.max_y))
        if len(self.wall_overlap(get_rect(coords))) > 0:
            return self.get_vacant_pos()
        else:
            return coords

    def get_thing_pos(self, props, typ):
        """Find a location for a thing with the given properties."""
        if "coords" in props:
            return self.get_cluster_pos(props["coords"][0], props["coords"][1])
        elif "cluster" in props:
            return self.get_cluster_pos(props["cluster"][1], self.get_cluster_center(typ))
        else:
            return self.get_vacant_pos()

    def get_cluster_center(self, typ):
        """Randomly select a cluster center."""
        return random.choice(self.clusters[typ])

    def get_cluster_pos(self, maxrad, center):
        """Find a position that doesn't overlap with any walls given a cluster center and radius."""
        coords = self.cluster(maxrad, center)
        if len(self.wall_overlap(get_rect(coords))) > 0:
            return self.get_cluster_pos(maxrad, center)
        else:
            return coords

    def cluster(self, maxrad, center):
        """Return a position given a cluster center and radius"""
        c_x = center[0]
        c_y = center[1]
        x = random.randint(0, maxrad)
        y = random.randint(0, int(math.sqrt(maxrad * maxrad - x * x)))
        if random.random() < .5:
            x = -x
        if random.random() < .5:
            y = -y
        return (max(self.min_x, min(self.max_x, c_x + x)),
                max(self.min_y, min(self.max_y, c_y + y)))

    def start_run(self, event):
        """Action for 'Run' button (2). Permits interrupting run.

        Thanks to David Bender."""
        if not self.running:
            self.running = True
            par = self.parent
            par.button0.config(state=DISABLED)
            par.button1.config(state=DISABLED)
            par.button2.config(text='Stop')
            par.button2.bind("<Button-1>", self.stop_run)
            if not par.showing_steps:
                 self.scan_dragto(100,100)
                 self.run_time = time.time()
            self.after_idle(self.run, event)

    def stop_run(self, event):
        """Interrupt run().

        Thanks to David Bender."""
        if self.running:
            print 'STOPPING RUN'
            self.abort_run = True

    def run(self, event):
        """Step the world # times indicated in Run length menu."""
        run_length = self.run_length()
        print '\nRUNNING WORLD FOR', run_length, 'STEPS,',
        if self.learn:
            print 'Q-LEARNING'
        elif self.reproduction == NO_REPRO:
            print 'NO LEARNING OR REPRODUCTION'
        elif self.reproduction == DUMB_COPYING:
            print 'COPYING WITHOUT EVOLUTION'
        elif self.reproduction == MATING:
            print 'EVOLUTION BY MATING'
        else:
            print 'EVOLUTION BY COPYING'
            
        for i in range(run_length):
            if self.abort_run:
                break
            if i % 100 == 0 and i > 0 and i != run_length - 1:
                self.run_stats(False)
            if self.parent.showing_steps:
                self.after(STEP_DELAY)
            self.__step(verbose = True)
            self.update()
        # Update thing graphics
        if not self.parent.showing_steps:
            self.update_graphics()
        self.run_stats(self.parent.showing_genomes)
        self.run_ended()

    def run_ended(self):
        """End run, possibly early.

        Thanks to David Bender."""
        if self.running:
#            print "FINISHED IN", time.time() - self.run_time
            self.running = False
            self.abort_run = False
            par = self.parent
            par.button0.config(state=NORMAL)
            par.button1.config(state=NORMAL)
            par.button2.config(text='Run')
            par.button2.bind("<Button-1>", self.start_run)
            self.scan_dragto(0,0)

    def run_stats(self, complete):
        '''Print useful statistics about the types in the population of orgs.'''
        print 'POPULATION AFTER', self.n_steps, 'STEPS'
        self.report_stats()
        for typ in self.features['THINGS'].iterkeys():
            if issubclass(typ, Org):
                strength_sum = 0.0
                age_sum = 0.0
                n = 0
                max_s = 0
                genomes = []
                for t1 in [t2 for t2 in self.things.itervalues() if isinstance(t2, typ)]:
                    if t1.has_genome():
                        genomes.append(t1.genome)
                    strength = t1.strength
                    strength_sum += strength
                    if strength > max_s:
                        max_s = strength
                    age_sum += t1.age
                    n += 1
                if n != 0:
                    print typ.__name__ + ':  N', n, ' mean strength', int(strength_sum / n), ' max strength', max_s, 'mean age', int(age_sum / n)
                if genomes and complete:
                    for g in genomes:
#                        g.show_bits()
                        g.show()

    def report_stats(self):
        if self.starved:
            print self.starved, 'critters died of weakness.'
            self.starved = 0
        if self.died_age:
            print self.died_age, 'critters died of old age.'
            self.died_age = 0
        if self.copied:
            print self.copied, 'critters copied themselves.'
            self.copied = 0
        if self.mated:
            print self.mated, 'pairs of critters mated.'
            self.mated = 0

    def __step(self, verbose = True):
        """Step the world once, verbosely or not."""
        self.regenerate()
        for t in self.things.itervalues():
            t.step(verbose = verbose)
        for s in self.sounds:
            s.step(verbose = verbose)
        self.respond_to_tags((("kill", "die"), ("chewed", "get_chewed")))
        self.reproduce()
        self.elim_events()
        self.n_steps += 1

    def step(self, event):
        """Step the world once."""
        self.__step()

    def elim_events(self):
        '''Get rid of all of the temporary events (like sounds).'''
        for s in self.sounds:
            if s.kill:
                self.sounds.remove(s)

    def update_graphics(self):
        '''Update graphics for things that have changed.'''
        for t in self.things.itervalues():
            t.update_graphics()

    def get_thing_n(self, kind_string):
        """Current number of things of the given kind."""
        return len(self.find_withtag(kind_string))

    def wall_overlap(self, crds):
        """List of all walls that overlap with coordinates crds."""
        found = []
        for o in self.find_overlapping(crds[0], crds[1], crds[2], crds[3]):
            if "wall" in self.gettags(o):
                c = self.coords(o)
                found.append((c[0] - WALL_THICKNESS / 2,
                              c[0] + WALL_THICKNESS / 2))
        return found

    def get_thing_seq(self, kind):
        '''List of things of type kind.'''
        return [item for item in self.things.values() if isinstance(item, kind)]

    def respond_to_tags(self, tags):
        """For each pair in tags, have all things with the tag call the action."""
        for t in tags:
            self.respond_to_tag(t[0], t[1])

    def respond_to_tag(self, tag, action):
        """All things with tag call method action (a string)."""
        for t in self.find_withtag(tag):
            thing = self.things[t]
            thing.dtag(tag)
            getattr(thing, action)()

    def toggle_show_senses(self, show):
        '''Toggle whether senses are displayed graphically.'''
        self.show_senses = show
        for c in self.get_thing_seq(Critter):
            for s in c.senses:
                s.toggle_show(show)

    def tag_overlapping(self, thing, tag, exclude = None, include = None, n = 1000):
        '''Tag things with tag that overlap with thing; return all tagged.'''
        rect = thing.get_rect()
        tagged = []
        if (include != None):
            ntagged = 0
            overlapping = self.find_overlapping(rect[0], rect[1], rect[2], rect[3])
            for o in overlapping:
                tags = self.gettags(o)
                if include in tags and "hidden" not in tags and ntagged < n:
                    self.addtag_withtag(tag, o)
                    tagged.append(o)
                    ntagged += 1
        else:
            id = thing.id
            tmp = "overlap" + str(id)
            self.addtag_overlapping(tmp, rect[0], rect[1], rect[2], rect[3])
            self.dtag(id, tmp)
            for t in exclude:
                self.dtag(t, tmp)
            tagged = self.find_withtag(tmp)
            self.addtag_withtag(tag, tmp)
            self.dtag("all", tmp)

        return tagged

    def overlapping(self, thing_id=None, rect=[], thing_label=None, exclude=''):
        '''All things overlapping with thing_id object or rect that have thing_label.'''
        rect = rect if rect else self.coords(thing_id)
        ol = []
        for tid in self.find_overlapping(rect[0], rect[1], rect[2], rect[3]):
            if tid != thing_id and tid in self.things and \
               (not thing_label or thing_label in self.gettags(tid)) \
               and (not exclude in self.gettags(tid)):
                ol.append(self.things[tid])
        return ol

class Wall:
    """Solid things that Orgs can't pass through."""

    def __init__(self, parent, vertical, pos = .5, start = 0, end = 1, exact = None,
                 world_width = WIDTH, world_height = HEIGHT, w = WALL_THICKNESS):
        '''Set the coordinates of the wall.'''
        self.world = parent
        self.vertical = vertical
        self.color = BRICK
        self.width = w

        if exact:
            self.pos = exact[0]
            self.start = exact[1]
            self.end = exact[2]
        elif vertical:
            self.pos = pos * world_width
            self.start = start * world_height
            self.end = end * world_height
        else:
            self.pos = pos * world_height
            self.start = start * world_width
            self.end = end * world_width

        if vertical:
            self.rect = (self.pos - self.width / 2, self.start,
                         self.pos + self.width / 2, self.end)
            self.id = self.world.create_line(self.pos, self.start, self.pos, self.end,
                                              width = self.width, fill = self.color)
        else:
            self.rect = (self.start, self.pos - self.width / 2,
                         self.end, self.pos + self.width / 2)
            self.id = self.world.create_line(self.start, self.pos, self.end, self.pos,
                                              width = self.width, fill = self.color)
            
        self.world.addtag_withtag("wall", self.id)
        self.world.addtag_withtag("obj", self.id)

    def get_bounds(self, plus_thing = True):
        '''Get the wall boundaries.'''
        incr = THING_SIDE if plus_thing else 0
        if self.vertical:
            return (self.pos - self.width / 2 - incr,
                    self.start - incr,
                    self.pos + self.width / 2 + incr,
                    self.end + incr)
        else:
            return (self.start - incr,
                    self.pos - self.width / 2 - incr,
                    self.end + incr,
                    self.pos + self.width / 2 + incr)

    def overlaps(self, stuff):
        '''Does thing/sense overlap with this wall?'''
        return stuff.id in self.world.find_overlapping(self.rect[0], self.rect[1],
                                                        self.rect[2], self.rect[3])

    def __str__(self):
        '''Print name for the wall.'''
        return "wall" + str(self.id)
